# ensure loading of all classes in metaclass
from .compound import *
from .full_text import *
from .geo import *
from .joining import *
from .shape import *
from .span import *
from .specialized import *
from .specialized_compound import *
from .term_level import *
