# your cluster address, if local probably 'localhost:9200'
ES_HOST = 'localhost:9300'

# where your table dumps are stored, and where serialized output will be written
DATA_DIR = '/Users/leonard/Alkemics/studies/initiatives/pandagg/data/imdb'
OUTPUT_FILE_NAME = 'serialized.json'
